unit URules;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils, fgl, DOM, XmlRead, XmlWrite, UXMLUtils;

type
  TRule = class;
  TRules = class;
  TGrammer = class;

  TRuleItemType = (ritTerminal, ritNonTerminal);

  { TRuleItem }

  TRuleItem = class
    ParentRule: TRule;
    RuleItemType: TRuleItemType;
    Terminal: string;
    NonTerminalName: string;
    NonTerminal: TRule;
    Grammer: TGrammer;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
    function GetBNF: string;
  end;

  TRuleType = (rtAnd, rtOr);

  { TRuleItems }

  TRuleItems = class(TFPGObjectList<TRuleItem>)
    ParentRule: TRule;
    Grammer: TGrammer;
    RuleType: TRuleType;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
    function GetBNF(RuleType: TRuleType): string;
  end;

  TRuleLinkType = (ltRequire);

  TRuleLink = class
    LinkType: TRuleLinkType;
    Rule: TRule;
  end;

  TRuleLinks = class(TFPGObjectList<TRuleLink>)

  end;

  { TRule }

  TRule = class
  private
    function GetGrammer: TGrammer;
    procedure SetGrammer(AValue: TGrammer);
  public
    Name: string;
    Items: TRuleItems;
    Links: TRuleLinks;
    property Grammer: TGrammer read GetGrammer write SetGrammer;
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
    function GetBNF: string;
  end;

  { TRules }

  TRules = class(TFPGObjectList<TRule>)
    Grammer: TGrammer;
    function FindName(Name: string): TRule;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
    function GetBNF: string;
  end;

  { TGrammer }

  TGrammer = class
    FileName: string;
    Rules: TRules;
    procedure LoadFromXmlFile(FileName: string);
    procedure SaveToXmlFile(FileName: string);
    constructor Create;
    destructor Destroy; override;
    function GetBNF: string;
  end;

implementation

{ TGrammer }

procedure TGrammer.LoadFromXmlFile(FileName: string);
var
  Doc: TXMLDocument;
  RootNode: TDOMNode;
  RulesNode: TDOMNode;
  Rule: TRule;
  RuleItem: TRuleItem;
begin
  Self.FileName := FileName;
  Rules.Clear;
  try
    ReadXMLFile(Doc, FileName);

    RootNode := Doc.DocumentElement;
    if RootNode.NodeName = 'GrammerProject' then begin
      RulesNode := RootNode.FindNode('Rules');
      if Assigned(RulesNode) then begin
        Rules.LoadFromXmlNode(RulesNode);
      end;
    end;
  finally
    Doc.Free;
  end;

  // Update rule references
  for Rule in Rules do
    for RuleItem in Rule.Items do
      if RuleItem.RuleItemType = ritNonTerminal then
        RuleItem.NonTerminal := Rules.FindName(RuleItem.NonTerminalName);
end;

procedure TGrammer.SaveToXmlFile(FileName: string);
var
  Doc: TXMLDocument;
  RootNode: TDOMNode;
  RulesNode: TDOMNode;
begin
  Self.FileName := FileName;
  Doc := TXMLDocument.Create;
  try
    RootNode := Doc.CreateElement('GrammerProject');
    Doc.AppendChild(RootNode);

    RulesNode := Doc.CreateElement('Rules');
    RootNode.AppendChild(RulesNode);
    Rules.SaveToXmlNode(RulesNode);

    WriteXMLFile(Doc, FileName);
  finally
    Doc.Free;
  end;
end;

constructor TGrammer.Create;
begin
  Rules := TRules.Create;
  Rules.Grammer := Self;
end;

destructor TGrammer.Destroy;
begin
  FreeAndNil(Rules);
  inherited Destroy;
end;

function TGrammer.GetBNF: string;
begin
  Result := Rules.GetBNF;
end;

{ TRuleItem }

procedure TRuleItem.SaveToXmlNode(Node: TDOMNode);
begin
  WriteInteger(Node, 'Type', Integer(RuleItemType));
  if RuleItemType = ritTerminal then WriteString(Node, 'Terminal', Terminal)
  else if RuleItemType = ritNonTerminal then WriteString(Node, 'NonTerminal', NonTerminal.Name)
end;

function TRuleItem.GetBNF: string;
begin
  if RuleItemType = ritTerminal then Result := '"' + Terminal + '"'
    else if RuleItemType = ritNonTerminal then Result := NonTerminal.Name;
end;

procedure TRuleItem.LoadFromXmlNode(Node: TDOMNode);
begin
  RuleItemType := TRuleItemType(ReadInteger(Node, 'Type', Integer(ritTerminal)));
  if RuleItemType = ritTerminal then Terminal := ReadString(Node, 'Terminal', '')
  else if RuleItemType = ritNonTerminal then
    NonTerminalName := ReadString(Node, 'NonTerminal', '');
end;

{ TRuleItems }

procedure TRuleItems.SaveToXmlNode(Node: TDOMNode);
var
  RuleItem: TRuleItem;
  RuleItemNode: TDOMNode;
begin
  WriteInteger(Node, 'Type', Integer(RuleType));

  for RuleItem in Self do begin
    RuleItemNode := Node.OwnerDocument.CreateElement('RuleItem');
    RuleItem.SaveToXmlNode(RuleItemNode);
    Node.AppendChild(RuleItemNode);
  end;
end;

function TRuleItems.GetBNF(RuleType: TRuleType): string;
var
  Item: TRuleItem;
begin
  Result := '';
  for Item in Self do begin
    if Item <> First then begin
      if RuleType = rtAnd then Result := Result + ' '
      else if RuleType = rtOr then Result := Result + ' | ';
    end;
    Result := Result + Item.GetBNF;
  end;
end;

procedure TRuleItems.LoadFromXmlNode(Node: TDOMNode);
var
  RuleItemNode: TDOMNode;
  RuleItem: TRuleItem;
begin
  RuleType := TRuleType(ReadInteger(Node, 'Type', Integer(rtAnd)));

  RuleItemNode := Node.FirstChild;
  while Assigned(RuleItemNode) do begin
    if RuleItemNode.NodeName = 'RuleItem' then begin
      RuleItem := TRuleItem.Create;
      RuleItem.Grammer := Grammer;
      RuleItem.LoadFromXmlNode(RuleItemNode);
      Add(RuleItem);
    end;
    RuleItemNode := RuleItemNode.NextSibling;
  end;
end;

{ TRule }

procedure TRule.SaveToXmlNode(Node: TDOMNode);
var
  RuleItemsNode: TDOMNode;
begin
  WriteString(Node, 'Name', Name);

  RuleItemsNode := Node.OwnerDocument.CreateElement('RuleItems');
  Node.AppendChild(RuleItemsNode);
  Items.SaveToXmlNode(RuleItemsNode);
end;

function TRule.GetBNF: string;
begin
  Result := Name + ' ::= ' + Items.GetBNF(Items.RuleType);
end;

function TRule.GetGrammer: TGrammer;
begin
  Result := Items.Grammer;
end;

procedure TRule.SetGrammer(AValue: TGrammer);
begin
  Items.Grammer := AValue;
end;

constructor TRule.Create;
begin
  Items := TRuleItems.Create;
  Items.ParentRule := Self;
  Links := TRuleLinks.Create;
end;

destructor TRule.Destroy;
begin
  FreeAndNil(Links);
  FreeAndNil(Items);
  inherited Destroy;
end;

procedure TRule.LoadFromXmlNode(Node: TDOMNode);
var
  ItemsNode: TDOMNode;
begin
  Name := ReadString(Node, 'Name', '');

  ItemsNode := Node.FindNode('RuleItems');
  if Assigned(ItemsNode) then begin
    Items.LoadFromXmlNode(ItemsNode);
  end;
end;

{ TRules }

procedure TRules.SaveToXmlNode(Node: TDOMNode);
var
  Rule: TRule;
  RuleNode: TDOMNode;
begin
  for Rule in Self do begin
    RuleNode := Node.OwnerDocument.CreateElement('Rule');
    Rule.SaveToXmlNode(RuleNode);
    Node.AppendChild(RuleNode);
  end;
end;

function TRules.FindName(Name: string): TRule;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (Items[I].Name <> Name) do Inc(I);
  if I < Count then Result := Items[I]
    else Result := nil;
end;

procedure TRules.LoadFromXmlNode(Node: TDOMNode);
var
  RuleNode: TDOMNode;
  Rule: TRule;
begin
  RuleNode := Node.FirstChild;
  while Assigned(RuleNode) do begin
    if RuleNode.NodeName = 'Rule' then begin
      Rule := TRule.Create;
      Rule.Grammer := Grammer;
      Rule.LoadFromXmlNode(RuleNode);
      Add(Rule);
    end;
    RuleNode := RuleNode.NextSibling;
  end;
end;

function TRules.GetBNF: string;
var
  Rule: TRule;
begin
  Result := '';
  for Rule in Self do begin
    Result := Result + Rule.GetBNF + LineEnding;
  end;
end;

end.

